"""Wrapper around prompt toolkit."""

import logging
from datetime import date, datetime
from typing import List

from prompt_toolkit import prompt as prmpt
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.validation import ValidationError, Validator


class DateValidator(Validator):
    """Date input validator."""

    def validate(self, document):
        """Validate input.

        Arguments:
            document -- Input stream to validate.

        Raises:
            ValidationError: When the number is not a digit.
            ValidationError: When the length is > 8 digits.
        """
        text = document.text

        if text and not text.isdigit():
            i = 0

            # Get index of fist non numeric character.
            # We want to move the cursor here.
            for i, value in enumerate(text):
                if not value.isdigit():
                    break

            raise ValidationError(message='Input is not a number',
                                  cursor_position=i)
        if text and not len(text) == 8:
            i = 0

            # Get index of fist non numeric character.
            # We want to move the cursor here.
            for i, value in enumerate(text):
                if not value.isdigit():
                    break

            raise ValidationError(message='This input is not 8 digits long.',
                                  cursor_position=i)


class NumberValidator(Validator):
    """Number input validation."""

    def validate(self, document):
        """Validate the input stream.

        Arguments:
            document -- Document input stream to validate.

        Raises:
            ValidationError: When input is not a number.
        """
        text = document.text

        if text and not text.isdigit():
            i = 0

            # Get index of fist non numeric character.
            # We want to move the cursor here.
            for i, value in enumerate(text):
                if not value.isdigit():
                    break

            raise ValidationError(message='Input is not a number',
                                  cursor_position=i)


class YesNoValidator(Validator):
    """Yes and No validation."""

    def validate(self, document):
        """Validate input strem.

        Arguments:
            document -- Document input to validate.

        Raises:
            ValidationError: When input is not a y or n.
        """
        text = document.text

        if text and text not in ["y", "Y", "n", "N"]:
            i = 0

            # Get index of fist non numeric character.
            # We want to move the cursor here.
            for i, value in enumerate(text):
                if not value.isdigit():
                    break

            raise ValidationError(message='Value not y or n',
                                  cursor_position=i)


def prompt(message: str, validator=None, completion=None, default="") -> str:
    """Request user for input.

    Arguments:
        message {str} -- Message for user.

    Keyword Arguments:
        validator {list} -- List of valid values. (default: {None})
        completion {list} -- List of completion options. (default: {None})

    Returns:
        str -- Response from user.
    """
    complete = WordCompleter(completion) if completion else None
    response = prmpt("{} >".format(message), validator=validator,
                     completer=complete, search_ignore_case=True, default=default)
    logging.info("%s: %s", message, response)
    return response


def prompt_date(message: str, default=None) -> date:
    """Request date from user.

    Arguments:
        message {str} -- Message to user.

    Returns:
        datetime -- Date that user input.
    """
    if not default:
        default = datetime.now().date().strftime("%d%m%Y")
    response = prompt("{} [DDMMYYYY]".format(message), validator=DateValidator(), default=default)
    form_date = datetime.strptime(response, "%d%m%Y").date()
    return form_date


def prompt_number(message: str) -> int:
    """Request number from user.

    Arguments:
        message {str} -- Message to user.

    Returns:
        int -- Number that user input.
    """
    num = int(prompt(message, validator=NumberValidator()))
    logging.info("%s: %s", message, num)
    return num


def prompt_yes_no(message: str) -> bool:
    """Request a yes or no from user.

    Arguments:
        message {str} -- Message to user.

    Returns:
        bool -- Response from user.
    """
    response = prompt("{} [y/n]".format(message), validator=YesNoValidator())
    return response in ['y', 'Y']


def prompt_option(message: str, options: List[str]) -> str:
    """Request a choice from a list of options

    Args:
        message (str): message to user.
        options (List): Options to provide in completion and validation

    Returns:
        str: [description]
    """
    def is_in_list(text):
        return text in options

    validator = Validator.from_callable(is_in_list,
                                        error_message="Input is not in list of options",
                                        move_cursor_to_end=True)
    response = prompt(message, completion=options, validator=validator)
    return response



"""General FarmOS Asset."""

from __future__ import annotations

from datetime import datetime
from typing import Dict, List, Optional, Union

from farmos_ext.farmobj import FarmObj
from farmos_ext.term import Crop, Season


class Asset(FarmObj):

    def __init__(self, farm, keys):
        if 'resource' not in keys:
            super().__init__(farm, keys)
        elif 'resource' in keys and keys['resource'] == 'farm_asset':
            super().__init__(
                farm, farm.asset.get({'id': keys['id']})['list'][0])

    @property
    def id(self) -> Optional[int]:  # pylint: disable=invalid-name
        return self.attr('id', int)

    @property
    def type(self) -> str:
        return self.attr('type', str)

    @property
    def description(self) -> Dict:
        return self.attr('description', str)

    @property
    def archived(self) -> Union[datetime, None]:
        key = self.key('archived')
        if key and key != '0':
            return FarmObj.timestamp_to_datetime(self.key('archived'))
        else:
            return None

    @property
    def flags(self) -> List[str]:
        return self.attr('flags', list)

    @property
    def created(self) -> Union[datetime, None]:
        return FarmObj.timestamp_to_datetime(self.key('created'))

    @property
    def changed(self) -> Union[datetime, None]:
        return FarmObj.timestamp_to_datetime(self.key('changed'))

    @property
    def uid(self) -> Union[int, None]:
        return int(self.key('uid')) if self.key('uid') else None

    @property
    def data(self) -> str:
        return self.attr('data', str)


class Planting(Asset):

    @property
    def crop(self) -> List[Crop]:
        return self.farm.terms(self.key('crop'), Crop)

    @property
    def season(self):
        return self.farm.terms(self.key('season'), Season)


class Animal(Asset):

    @property
    def animal_type(self) -> str:
        return self.attr('animal_type', str)

    @property
    def nicknames(self) -> List[str]:
        return self.attr('animal_nicknames', list)

    @property
    def castrated(self) -> bool:
        return self.attr('animal_castrated', bool)

    @property
    def sex(self) -> str:
        return self.attr('animal_sex', str)

    @property
    def tag(self):
        return self.attr('tag', str)

    @property
    def parent(self) -> List[Animal]:
        return self.farm.assets(self.key('parent'), Animal)

    @property
    def birth_date(self) -> Union[datetime, None]:
        return FarmObj.timestamp_to_datetime(self.key('date'))


class Equipment(Asset):

    @property
    def manufacturer(self) -> str:
        return self.attr('manufacturer', str)

    @property
    def model(self) -> str:
        return self.attr('model', str)

    @property
    def serial_number(self) -> str:
        return self.attr('serial_number', str)


class Sensor(Asset):
    pass


class Compost(Asset):
    pass




from dataclasses import dataclass
from enum import Enum
from typing import Dict, Union

from farmos_ext.farmobj import FarmObj


class Measure(Enum):
    COUNT = 'count'
    LENGTH = 'length'
    AREA = 'area'
    VOLUME = 'volume'
    TIME = 'time'
    TEMPERATURE = 'temperature'
    VALUE = 'value'
    RATE = 'rate'
    RATING = 'rating'
    RATIO = 'ratio'
    PROBABILITY = 'probability'
    WEIGHT = 'weight'


@dataclass
class Quantity():
    measure: Measure
    label: str
    unit: str
    value: str

    def to_dict(self):
        return {
            "measure": self.measure.value,
            "unit": {
                'name': self.unit,
                "resource": "taxonomy_term"
            },
            "value": self.value,
            "label": self.label
        }


@dataclass
class Inventory():
    value: int
    asset_id: int

    def to_dict(self):
        return {
            "asset": {"id": self.asset_id},
            "value": str(self.value)
        }


class User(FarmObj):
    pass


class Content(FarmObj):

    @property
    def api_version(self) -> Union[str, None]:
        return self.key('api_version')

    @property
    def system_of_measurement(self) -> Union[str, None]:
        return self.key('system_of_measurement')

    @property
    def metrics(self) -> Dict:
        return self.key('metrics')

    @property
    def mapbox_api_key(self) -> str:
        return self.key('mapbox_api_key')

    @property
    def languages(self) -> Dict:
        return self.key('languages')

    @property
    def google_maps_api_key(self) -> str:
        return self.key('google_maps_api_key')

    @property
    def resources(self) -> Dict:
        return self.key('resources')


class Soil(FarmObj):
    pass



from __future__ import annotations

from datetime import datetime, timedelta
from typing import Dict, Iterator, List, Optional

from farmos_ext.farmobj import FarmObj
from farmos_ext.file import File
from farmOS import farmOS  # pylint: disable=wrong-import-order


class Term(FarmObj):

    def __init__(self, farm: farmOS, keys: Dict):
        if 'resource' not in keys:
            super().__init__(farm, keys)
        elif 'resource' in keys and keys['resource'] == 'taxonomy_term':
            super().__init__(farm, farm.term.get({"tid": int(keys['id'])})['list'][0])
        else:
            raise KeyError('Key resource does not have value taxonomy_term')

    @property
    def tid(self) -> Optional[int]:
        return int(self._keys['tid']) if self._keys['tid'] else None

    @property
    def weight(self) -> Optional[int]:
        return int(self._keys['weight']) if self._keys['weight'] else None

    @property
    def description(self) -> str:
        return self.key('description')

    @property
    def parent(self) -> List[Term]:
        return self.farm.terms(self._keys['parent'], Term)

    @property
    def vocabulary(self) -> Optional[Dict]:
        return self.key('vocabulary')


class Season(Term):

    @property
    def start_date(self) -> Optional[datetime]:
        return FarmObj.timestamp_to_datetime(self._keys['date_range']['value'])

    @property
    def end_date(self) -> Optional[datetime]:
        return FarmObj.timestamp_to_datetime(self._keys['date_range']['value2'])

    @property
    def duration(self) -> Optional[timedelta]:
        if self.start_date and self.end_date:
            return self.end_date - self.start_date
        return None


class CropFamily(Term):
    pass


class Crop(Term):

    @property
    def companions(self) -> List[Crop]:
        return self.farm.terms(self._keys['companions'], Crop)

    @property
    def crop_family(self) -> Optional[CropFamily]:
        key = self._keys['crop_family']
        return CropFamily(self._farm, key) if key else None

    @property
    def maturity_days(self) -> Optional[int]:
        key = self._keys['maturity_days']
        return int(self._keys['maturity_days']) if key else None

    @property
    def parents_all(self) -> List[Crop]:
        return self.farm.terms(self._keys['parents_all'], Crop)

    @property
    def transplant_days(self) -> Optional[int]:
        key = self.key('transplant_days')
        return int(key) if key else None

    @property
    def images(self) -> Iterator[File]:
        keys = self.key('images')
        if keys:
            for key in keys:
                if 'file' in key:
                    yield File(self.farm, key['file'])


class Unit(FarmObj):
    pass


class Category(FarmObj):
    pass



"""Logging output."""

import logging

import colorama

colorama.init()


def init(name: str):
    """Initialize output logging."""
    logging.basicConfig(filename=name, filemode='w', level=logging.INFO,
                        format='%(name)s - %(levelname)s - %(message)s')


def _message(msg: str, color):
    print(color, end="")
    print(msg)
    print(colorama.Fore.RESET, end="")


def message(msg: str, level=0, color=colorama.Fore.WHITE):
    """Create a basic message. Level Info."""
    _message("{}{}".format("\t" * level, msg), color)
    logging.info(msg)


def info(msg: str):
    """Create info level log."""
    logging.info(msg)


def alert(msg: str):
    """Create warning level log."""
    _message(msg, colorama.Fore.YELLOW)
    logging.warning(msg)


def error(msg: str):
    """Create error level log."""
    _message(msg, colorama.Fore.RED)
    logging.error(msg)


def debug(msg: str):
    """Create a debug log."""
    _message(msg, colorama.Fore.GREEN)
    logging.debug(msg)



"""FarmOS Area objects."""

from __future__ import annotations

from typing import Dict, List, Union

from farmos_ext.asset import Asset
from farmos_ext.farmobj import FarmObj


class Area(FarmObj):
    """Area located within a Farm."""

    @property
    def assets(self) -> List[Asset]:
        """Assets assigned to area.

        Returns:
            Union[List[Asset], None]: [description]
        """
        return self.farm.assets(self.key('assets'))

    @property
    def description(self) -> str:
        """Explanation string of the asset.

        This is not escaped for any formatting that is present.

        Returns:
            str: Unformatted string.
        """
        return self.key('description')

    @property
    def flags(self) -> List[str]:
        """Flag items assigned.

        Returns:
            List[str]: Assigned flags by name.
        """
        return self.key('flags')

    @property
    def geofield(self) -> List[Dict]:
        """Location of area in `Well-Known Text` format.

        Returns:
            List[Dict]: Described locations.
        """
        return self.key('geofield')

    @property
    def parent(self) -> List[Area]:
        """Parent Area of this Area.

        Returns:
            List[Area]: Generally a single item but stored as a list.
        """
        return self.farm.areas(self.key('parent'))

    @property
    def parents_all(self) -> List[Area]:
        """All parents.

        Returns:
            List[Area]: List of parents.
        """
        return self.farm.areas(self.key('parents_all'))

    @property
    def tid(self) -> Union[int, None]:
        """Taxonomy ID of item.

        Returns:
            Union[int, None]: Unique ID, none if there is no id but this should
            not occur.
        """
        return self.attr('tid', int)

    @property
    def vocabulary(self) -> Dict:
        """Vocabulary item assigned.

        Returns:
            Dict: Vocabulary item with id and name.
        """
        return self.key('vocabulary')





import json
import time
from datetime import datetime
from typing import Dict, List
import os

import requests

FULL_URL = "{0}{1}{2}"
SENSOR_URL = "farm/sensor/listener/"


class FarmSensorException(Exception):
    pass


class Sensor():

    def __init__(self, farm: str = None, pub: str = None, prv: str = None) -> None:
        if not pub:
            if os.environ.get("FARM_SENSOR_PUB_KEY"):
                self._public_key = os.environ.get("FARM_SENSOR_PUB_KEY")
            else:
                raise FarmSensorException("No Sensor Public Key found")
        else:
            self._public_key = pub

        if not prv:
            if os.environ.get("FARM_SENSOR_PRV_KEY"):
                self._private_key = os.environ.get("FARM_SENSOR_PRV_KEY")
        else:
            self._private_key = prv

        if not farm:
            if os.environ.get('FARM_SENSOR_HOST'):
                farm = os.environ.get('FARM_SENSOR_HOST')
            else:
                raise FarmSensorException("No Farm Host URL found")
        self._url = FULL_URL.format(farm, SENSOR_URL, self._public_key)

    def upload(self, data: Dict, sleep=True):
        requests.post(self._url, params={"private_key": self._private_key}, json=data)
        if sleep:
            countdown = "Upload delay ending in: {} seconds"
            for index in range(0, 70):
                print(countdown.format(70-index), end="\r")
                time.sleep(1)
            print(" " * len(countdown), end='\r')

    def get(self, name: str = None, start: datetime = None, end: datetime = None, limit: int = 0) -> List[Dict]:
        data = {}
        data['limit'] = limit
        if name:
            data['name'] = name
        if start:
            data['start'] = start.timestamp()
        if end:
            data['end'] = end.timestamp()
        if self._private_key:
            data['private_key'] = self._private_key
        req = requests.get(self._url, params=data)
        record = json.loads(req.text)
        return record

    def summary(self) -> Dict:
        return json.loads(requests.get("{}/summary".format(self._url), params={"private_key": self._private_key}).text)




from __future__ import annotations

from datetime import datetime
import os

from typing import Dict, Optional, TYPE_CHECKING

from farmos_ext.farmobj import FarmObj

if TYPE_CHECKING:
    from farmos_ext.farm import Farm  # pylint: disable=cyclic-import


class File(FarmObj):

    def __init__(self, farm: Farm, keys: Dict = None):
        if 'resource' not in keys:
            super().__init__(farm, keys)
        elif 'resource' in keys and keys['resource'] == 'file':
            super().__init__(
                farm, farm.file.get({'id': keys['id']})['list'][0])
        else:
            raise KeyError('Key resource does not have value log')

    @property
    def mime(self) -> Optional[str]:
        return self.attr('mime', str)

    @property
    def size(self) -> Optional[int]:
        return self.attr('size', int)

    @property
    def url(self) -> Optional[str]:
        return self.attr('url', str)

    @property
    def id(self) -> Optional[int]:  # pylint: disable=invalid-name
        return self.attr('fid', int)

    @property
    def timestamp(self) -> Optional[datetime]:
        return FarmObj.timestamp_to_datetime(self.attr('timestamp', int))

    @property
    def owner(self):
        return None

    def download(self):
        data = self.farm.session.get(self.url)
        path = os.path.join(self.farm.local_resources, 'files', self.name)
        open(path, 'wb').write(data.content)
        return path




from __future__ import annotations

from typing import Dict, List, Union

from farmos_ext.reporting.report import Report

TITLE = "="

HEADING = [
    TITLE,
    "=",
    "-",
    "^",
    "*"
]


class RstReporter(Report):

    def __init__(self, filename: str):
        self._doc = ""
        super().__init__(filename)
        self.title(self.filename)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.save()

    @staticmethod
    def _sanitize(text: str):
        cleaned = text.replace("<p>", "").replace("</p>", "").replace("&nbsp;", "\n").replace("\n\n", "\n").replace(
            "<br />", "\n\n").replace("<dl>", "\n").replace("</dl>", "\n").replace("<dd>", "\n\n").replace(
                "</dd>", "\n\n").replace("<dt>", "**").replace("</dt>", "**\n")
        return cleaned.replace("\n\n", "\n")

    def _append(self, text: str):
        self._doc += RstReporter._sanitize(text)

    def line(self, text=""):
        if text is not None:
            self._append("{}\n".format(text))

    def directive(self, name: Union[str, Dict[str, str]],
                  configs: Union[None, List[Union[str, Dict[str, str]]]]):
        self.line()
        if isinstance(name, str):
            self.line(".. {}::".format(name))
        elif isinstance(name, dict):
            for key in name:
                self.line(".. {}:: {}".format(key, name[key]))
        if configs:
            for config in configs:
                if isinstance(config, str):
                    self.line("    :{}:".format(config))
                elif isinstance(config, dict):
                    for key in config:
                        self.line("    :{}: {}".format(key, config[key]))
        self.line()

    def title(self, text: str):
        self.line(TITLE * len(text))
        self.line(text)
        self.line(TITLE * len(text))
        self.line()

    def heading(self, text: str, level: int):
        self.line()
        self.line(text)
        self.line(HEADING[level] * len(text))
        self.line()

    def lists(self, items: List[str], ordered=True):
        denote = "#" if ordered else "-"
        self.line()
        for item in items:
            self.line("{} {}".format(denote, item))
        self.line()

    def definition(self, key: str, term: str):
        self.line(":{}: {}".format(key, term))

    def toctree(self, depth=1):
        self.directive("contents", ["local", {"depth": depth}])

    def pagebreak(self):
        self.directive({"raw": "pdf"}, None)
        self.line("    PageBreak")
        self.line()

    def table(self, items: List[Dict]):
        def row(cols: List[str], widths: List[int]):
            new_list = []
            for index, col in enumerate(cols):
                new_list.append("{col:{width}}".format(
                    col=col, width=widths[index]))
            return "| {} |".format(" | ".join(new_list))

        def sep(widths: List[int]):
            self.line(
                "+{}+".format("+".join(["-"*(index+2) for index in widths])))

        keys = list(items[0].keys())
        col_widths = [0] * len(keys)
        for index, key in enumerate(keys):
            if len(key) > col_widths[index]:
                col_widths[index] = len(key)
            for item in items:
                if len(item[key]) > col_widths[index]:
                    col_widths[index] = len(item[key])
        header = row(keys, col_widths)
        sep(col_widths)
        self.line(header)
        sep(col_widths)

        for item in items:
            self.line(row([item[key] for key in keys], col_widths))
            sep(col_widths)

    def image(self, path):
        self.directive({'image': path}, None)

    def save(self):
        path = "{}.rst".format(self.filename) if not self.filename.endswith(
            ".rst") else self.filename
        with open(path, 'w', encoding='utf-8') as rst:
            rst.write(self._doc)




from farmos_ext.farmobj import FarmObj
from farmos_ext.term import Season


class Plan(FarmObj):
    pass


class CropPlan(Plan):

    @property
    def notes(self) -> str:
        return self.attr('notes', str)

    @property
    def season(self) -> Season:
        return self.attr('season', Season)


class GrazingPlan(Plan):

    @property
    def field_days_bulk_feeding(self) -> int:
        return self.attr('field_days_bulk_feeding', int)

    @property
    def field_days_of_drought_reserve(self) -> int:
        return self.attr('field_days_of_drought_reserve', int)

    @property
    def field_expected_rotations(self) -> int:
        return self.attr('field_expected_rotations', int)

    @property
    def grazing_factors(self) -> str:
        return self.attr('grazing_factors', str)

    @property
    def field_grazing_growing_season(self) -> bool:
        return self.attr('field_grazing_growing_season', bool)

    @property
    def date_range(self):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['farmos_ext', 'farmos_ext.cli', 'farmos_ext.reporting']

package_data = \
{'': ['*']}

install_requires = \
['colorama>=0.4.4,<0.5.0',
 'farmOS>=0.2.0,<0.3.0',
 'prompt-toolkit>=3.0.16,<4.0.0']

setup_kwargs = {
    'name': 'farmos-ext',
    'version': '0.1.0',
    'description': '',
    'long_description': None,
    'author': 'Lucas Brendel',
    'author_email': 'lucasbrendel@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)




from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, Optional, Type

if TYPE_CHECKING:
    from farmos_ext.farm import Farm  # pylint: disable=cyclic-import


class FarmObj():

    _farm = None

    def __init__(self, farm: Farm, keys: Dict = None):
        self._farm = farm
        self._keys = keys

    @property
    def farm(self):
        return self._farm

    @staticmethod
    def _basic_prop(prop: Any) -> Any:
        """Returns the argument if it is not None."""
        return prop if prop else None

    @staticmethod
    def timestamp_to_datetime(timestamp: int) -> Optional[datetime]:
        """Convert timestampt to datetime or return None."""
        return datetime.fromtimestamp(int(timestamp)) if timestamp else None

    def key(self, key: str) -> Optional[Any]:
        if key in self._keys:
            return self._keys[key]
        else:
            return None

    def attr(self, key: str, ret_type: Type[Any]) -> Type[Any]:
        value = FarmObj._basic_prop(self.key(key))
        return ret_type(value) if value else value

    @property
    def name(self) -> str:
        return self.key('name')



"""Main farm access."""

from __future__ import annotations

import os
from datetime import datetime
from typing import Dict, Iterable, Iterator, List, Type, Union

from farmos_ext.area import Area
from farmos_ext.asset import Asset, Equipment, Planting
from farmos_ext.log import (Activity, Birth, Harvest, Input, Log, Maintenance,
                            Medical, Observation, Purchase, Sale, Seeding,
                            SoilTest, Transplanting)
from farmos_ext.others import Content, Quantity
from farmos_ext.term import Crop, CropFamily, Season, Term, Unit
from farmOS import farmOS  # pylint: disable=wrong-import-order
from farmOS.client import BaseAPI  # pylint: disable=wrong-import-order


class FarmTypeMissingError(Exception):
    pass


def farm():
    """Access to farm with provided credentials."""
    return Farm()


class FileAPI(BaseAPI):

    def __init__(self, session):
        # Define 'log' as the farmOS API entity endpoint
        super().__init__(session=session, entity_type='file')


# pylint: disable=too-many-public-methods
class Farm(farmOS):

    def __init__(self, local_resources="./resources"):
        self._host = None
        self._user = None
        self._pass = None
        self.local_resources = local_resources

        if os.path.exists("farmos.cfg"):
            with open('farmos.cfg') as cfg:
                for line in cfg.readlines():
                    if line.startswith("HOST"):
                        self._host = line[line.index("=")+1:].strip()
                    if line.startswith("USER"):
                        self._user = line[line.index("=")+1:].strip()
                    if line.startswith("PASS"):
                        self._pass = line[line.index("=")+1:].strip()
            if not self._host:
                raise KeyError("HOST key is not defined in farmos.cfg")
            if not self._user:
                raise KeyError("USER key is not defined in farmos.cfg")
            if not self._pass:
                raise KeyError("PASS key is not defined in farmos.cfg")
            super().__init__(self._host)
            self._token = self.authorize(self._user, self._pass)
        else:
            raise Exception('farmos.cfg not found.')
        self.file = FileAPI(self.session)

    def assets(self,
               filters: Union[Dict, List[Dict], int, str] = None,
               asset_class: Type[Asset] = Asset) -> Iterator[Type[Asset]]:
        if isinstance(filters, list):
            for filt in filters:
                for asset in self.asset.get(filt)['list']:
                    yield asset_class(self, keys=asset)
        else:
            for asset in self.asset.get(filters)['list']:
                yield asset_class(self, keys=asset)

    # def _get_assets(self, items: List[Dict], obj_class):
    #     retitems = []
    #     for item in items:
    #         rets = self.asset.get(item['id'])
    #         if 'list' in rets:
    #             self.extract(rets, obj_class)
    #         else:
    #             retitems.append(obj_class(self, rets))
    #     return retitems

    def logs(self,
             filters: Union[Dict, List[Dict], int, str] = None,
             log_class: Type[Log] = Log) -> Iterator[Type[Log]]:
        if isinstance(filters, list):
            for filt in filters:
                for log in self.log.get(filt):
                    yield log_class(self, keys=log)
        elif isinstance(filters, int):
            yield log_class(self, keys=self.log.get(filters))
        else:
            for log in self.log.get(filters):
                yield log_class(self, keys=log)

    def terms(self, filters: Union[str, List[Dict], Dict] = None,
              term_class: Type[Term] = Term) -> Iterator[Type[Term]]:
        if isinstance(filters, list):
            for item in filters:
                for term in self.term.get({"tid": item['id']})['list']:
                    yield term_class(self, keys=term)
        else:
            rets = self.term.get(filters)
            yield term_class(self, keys=rets)

    def areas(self, filters: Union[Dict, List[Dict], int, str] = None) -> Iterator[Area]:
        if isinstance(filters, list):
            for filt in filters:
                for area in self.area.get(filt)['list']:
                    yield Area(self, keys=area)
        else:
            for area in self.area.get(filters)['list']:
                yield Area(self, keys=area)

    def _create_log(self, name: str, date: datetime, category: str, fields: Dict, done=False):
        data = {
            "name": name,
            "timestamp": str(int(datetime.timestamp(date))),
            "log_category": [{
                "name": category
            }],
            "type": "farm_observation"
        }
        data.update(fields)
        if 'done' not in data:
            data['done'] = '1' if done else '0'
        ret = self.log.send(data)
        return ret

    @property
    def content(self) -> Content:
        return Content(self, keys=self.info())

    @property
    def seasons(self) -> Iterator[Season]:
        for season in self.term.get("farm_season")['list']:
            yield Season(self, season)

    @property
    def crop_families(self) -> Iterator[CropFamily]:
        for fam in self.term.get("farm_crop_families")['list']:
            yield CropFamily(self, keys=fam)

    @property
    def crops(self) -> Iterator[Crop]:
        for crop in self.term.get("farm_crops")['list']:
            yield Crop(self, crop)

    def equipment(self, filters: Dict = None) -> Iterable[Equipment]:
        if not filters:
            filters = {'type': 'equipment'}
        else:
            filters.update({'type': 'equipment'})
        return self.assets(filters, Equipment)

    def plantings(self, filters: Dict = None) -> Iterable[Planting]:
        if not filters:
            filters = {'type': 'planting'}
        else:
            filters.update({'type': 'planting'})
        return self.assets(filters, Planting)

    @property
    def units(self) -> Iterable[Unit]:
        for unit in self.term.get('farm_quantity_units')['list']:
            yield Unit(self, unit)

    def harvests(self, filters: Dict = None) -> Iterable[Harvest]:
        if 'farm_harvests' in self.content.resources['log']:
            if not filters:
                filters = {'type': 'farm_harvest'}
            else:
                filters.update({'type': 'farm_harvest'})
            return self.logs(filters, Harvest)
        else:
            raise FarmTypeMissingError("Harvest logs not supported.")

    def seedings(self, filters: Dict = None) -> Iterable[Seeding]:
        if 'farm_seedings' in self.content.resources['log']:
            if not filters:
                filters = {'type': 'farm_seeding'}
            else:
                filters.update({'type': 'farm_seeding'})
            return self.logs(filters, Seeding)
        else:
            raise FarmTypeMissingError("Seeding logs not supported.")

    def transplants(self, filters: Dict = None) -> Iterable[Transplanting]:
        if 'farm_transplanting' in self.content.resources['log']:
            if not filters:
                filters = {'type': 'farm_transplanting'}
            else:
                filters.update({'type': 'farm_transplanting'})
            return self.logs(filters, Transplanting)
        else:
            raise FarmTypeMissingError("Transplanting logs not supported.")

    def observations(self, filters: Dict = None) -> Iterable[Observation]:
        if 'farm_observation' in self.content.resources['log']:
            if not filters:
                filters = {'type': 'farm_observation'}
            else:
                filters.update({'type': 'farm_observation'})
            return self.logs(filters, Observation)
        else:
            raise FarmTypeMissingError("Observation logs not supported.")

    def maintenances(self, filters: Dict = None) -> Iterator[Maintenance]:
        if 'farm_maintenance' in self.content.resources['log']:
            if not filters:
                filters = {'type': 'farm_maintenance'}
            else:
                filters.update({'type': 'farm_maintenance'})
            return self.logs(filters, Maintenance)
        else:
            raise FarmTypeMissingError("Maintenance logs not supported.")

    def purchases(self, filters: Dict = None) -> Iterator[Purchase]:
        if 'farm_purchase' in self.content.resources['log']:
            if not filters:
                filters = {'type': 'farm_purchase'}
            else:
                filters.update({'type': 'farm_purchase'})
            return self.logs(filters, Purchase)
        else:
            raise FarmTypeMissingError("Purchase logs not supported.")

    def sales(self, filters: Dict = None) -> Iterator[Sale]:
        if 'farm_sale' in self.content.resources['log']:
            if not filters:
                filters = {'type': 'farm_sale'}
            else:
                filters.update({'type': 'farm_sale'})
            return self.logs(filters, Sale)
        else:
            raise FarmTypeMissingError("Sale logs not supported.")

    def births(self, filters: Dict = None) -> Iterator[Birth]:
        if 'farm_birth' in self.content.resources['log']:
            if not filters:
                filters = {'type': 'farm_birth'}
            else:
                filters.update({'type': 'farm_birth'})
            return self.logs(filters, Birth)
        else:
            raise FarmTypeMissingError("Birth logs not supported.")

    def inputs(self, filters: Dict = None) -> Iterator[Input]:
        if 'farm_input' in self.content.resources['input']:
            if not filters:
                filters = {'type': 'farm_input'}
            else:
                filters.update({'type': 'farm_input'})
            return self.logs(filters, Input)
        else:
            raise FarmTypeMissingError("Input logs not supported.")

    def soil_tests(self, filters: Dict = None) -> Iterator[SoilTest]:
        if 'farm_soil_test' in self.content.resources['log']:
            if not filters:
                filters = {'type': 'farm_soil_test'}
            else:
                filters.update({'type': 'farm_soil_test'})
            return self.logs(filters, SoilTest)
        else:
            raise FarmTypeMissingError("Soil test logs not supported.")

    def activities(self, filters: Dict = None) -> Iterator[Activity]:
        if 'farm_activity' in self.content.resources['log']:
            if not filters:
                filters = {'type': 'farm_activity'}
            else:
                filters.update({'type': 'farm_activity'})
            return self.logs(filters, Activity)
        else:
            raise FarmTypeMissingError("Activity logs not supported.")

    def medicals(self, filters: Dict = None) -> Iterator[Medical]:
        if 'farm_medical' in self.content.resources['log']:
            if not filters:
                filters = {'type': 'farm_medical'}
            else:
                filters.update({'type': 'farm_medical'})
            return self.logs(filters, Medical)
        else:
            raise FarmTypeMissingError("Medical logs are not supported.")

    def create_planting(self, crop: Crop, season: str, location: str) -> Planting:
        ret = self.asset.send({
            "name": "{} {} {}".format(season, location, crop.name),
            "type": "planting",
            "crop": [
                {
                    "id": crop.tid
                }
            ],
            "season": [{"name": season}]
        })
        plant = Planting(self, keys=ret)
        return plant

    def create_seeding(self, planting: Planting, location: Area, crop: Crop,
                       date: datetime, seeds: int, source=None, done=False) -> Seeding:
        name = "Seed {} {} {}".format(date.year, location.name, crop.name)
        fields = {
            "type": "farm_seeding",
            "asset": [
                {
                    "id": planting.id,
                    "resource": "taxonomy_term"
                }
            ],
            "seed_source": source,
            "movement": {
                "area": [
                    {
                        "id": location.tid,
                        "resource": "taxonomy_term"
                    }
                ]
            },
            "quantity": [
                {
                    "measure": "count",
                    "value": str(seeds),
                    "unit": {
                        'name': 'Seeds',
                        "resource": "taxonomy_term"
                    }
                }
            ]
        }
        ret = self._create_log(name, date, 'Plantings', fields, done=done)
        return Seeding(self, keys=ret)

    def create_transplant(self, planting: Planting, location: Area, date: datetime, fields=None, done=False):
        name = "Transplant {}".format(planting.name)
        data = {
            "type": "farm_transplanting",
            "movement": {
                "area": [
                    {
                        "id": location.tid,
                        "resource": "taxonomy_term"
                    }
                ]
            },
            "asset": [
                {
                    "id": planting.id,
                    "resource": "taxonomy_term"
                }
            ]
        }
        if fields:
            data.update(fields)
        ret = self._create_log(name, date, 'Plantings', data, done=done)
        return Transplanting(self, ret)

    def create_harvest(self, planting: Planting, date: datetime, quantities: List[Quantity], done=False):
        name = "Harvest {} {}".format(date.year, planting.crop[0]['name'])
        data = {
            "type": "farm_harvest",
            "asset": [{
                "id": planting.id,
                "resource": "taxonomy_term"
            }]
        }

        if quantities:
            data["quantity"] = []
            for quantity in quantities:
                data["quantity"].append(quantity.to_dict())

        ret = self._create_log(name, date, 'Plantings', data, done=done)
        return Harvest(self, ret)

    def create_log(self, name: str, date: datetime, category: str, fields: Dict, done=False):
        return Log(self, self._create_log(name, date, category, fields, done))




from __future__ import annotations

from datetime import datetime
from typing import Dict, Iterator, List, Optional

from farmos_ext.area import Area
from farmos_ext.asset import Asset, Equipment
from farmos_ext.farmobj import FarmObj
from farmos_ext.file import File
from farmos_ext.others import Inventory, Quantity, Soil
from farmos_ext.term import Category, Unit


# pylint: disable=too-many-public-methods
class Log(FarmObj):

    def __init__(self, farm, keys: Dict):
        if 'resource' not in keys:
            super().__init__(farm, keys)
        elif 'resource' in keys and keys['resource'] == 'log':
            super().__init__(
                farm, farm.log.get({'id': keys['id']})['list'][0])
        else:
            raise KeyError('Key resource does not have value log')

    @property
    def id(self) -> Optional[int]:  # pylint: disable=invalid-name
        return self.attr('id', int)

    @property
    def type(self) -> str:
        return self.attr('type', str)

    @property
    def timestamp(self) -> Optional[datetime]:
        return FarmObj.timestamp_to_datetime(self.key('timestamp'))

    @property
    def done(self) -> bool:
        return bool(self.attr('done', int))

    @property
    def notes(self) -> Optional[str]:
        return self.key('notes')['value'] if self.key('notes') else None

    @property
    def asset(self) -> List[Asset]:
        key = self.key('asset')
        return self.farm.assets(key, Asset) if key else []

    @property
    def equipment(self) -> List[Equipment]:
        key = self.key('equipment')
        return self.farm.assets(key, Equipment) if key else []

    @property
    def area(self) -> List[Area]:
        key = self.key('area')
        return self.farm.areas(key, Area) if key else []

    @property
    def geofield(self) -> Optional[str]:
        return self.attr('geofield', str)

    @property
    def movement(self) -> List[Area]:
        if self.key('movement'):
            key = self.key('movement')['area']
            return self.farm.areas(key, Area) if key else []
        return []

    @property
    def membership(self) -> str:
        return self.attr('membership', str)

    @property
    def quantity(self) -> List[Quantity]:
        if self.key('quantity'):
            ret = []
            for quantity in self.key('quantity'):
                ret.append(Quantity(measure=quantity['measure'],
                                    label=quantity['label'],
                                    value=quantity['value'],
                                    unit=quantity['unit'] if 'unit' in quantity else None))
            return ret
        return []

    @property
    def flags(self) -> List[str]:
        return self.attr('flags', list)

    @property
    def categories(self) -> List[Category]:
        key = self.key('log_category')
        return [Category(self._farm, x) for x in key] if key else []

    @property
    def owner(self):
        return self.attr('log_owner', str)

    @property
    def created(self) -> Optional[datetime]:
        return FarmObj.timestamp_to_datetime(self.key('created'))

    @property
    def changed(self) -> Optional[datetime]:
        return FarmObj.timestamp_to_datetime(self.key('changed'))

    @property
    def uid(self) -> Optional[int]:
        key = self.key('uid')
        return int(key) if key else None

    @property
    def data(self) -> str:
        return self.attr('data', str)

    @property
    def inventory(self) -> List[Inventory]:
        if self.key('inventory'):
            ret = []
            for inventory in self.key('inventory'):
                ret.append(Inventory(inventory['value'], inventory['asset']['id']))
            return ret
        return []

    @property
    def images(self) -> Iterator[File]:
        keys = self.key('images')
        if keys:
            for key in keys:
                if 'file' in key:
                    yield File(self, key['file'])

    @property
    def files(self) -> Iterator[File]:
        keys = self.key('files')
        if keys:
            for key in keys:
                if 'file' in key:
                    yield File(self, key['file'])


class LotLog(Log):

    @property
    def lot_number(self) -> str:
        return self.attr('lot_number', str)


class MoneyLog(LotLog):

    @property
    def units(self) -> List[Unit]:
        return []

    @property
    def values(self) -> List:
        return []

    @property
    def total_price(self) -> float:
        return self.attr('total_price', float)

    @property
    def unit_price(self) -> float:
        return self.attr('unit_price', float)


class Input(LotLog):

    @property
    def material(self) -> str:
        return self.attr('material', str)

    @property
    def purpose(self) -> str:
        return self.attr('input_purpose', str)

    @property
    def method(self) -> str:
        return self.attr('input_method', str)

    @property
    def source(self) -> str:
        return self.attr('input_source', str)

    @property
    def date_purchase(self) -> Optional[datetime]:
        return FarmObj.timestamp_to_datetime(self.key('date_purchase'))


class Seeding(LotLog):

    @property
    def seed_source(self) -> str:
        return self.attr('seed_source', str)


class Transplanting(Log):
    pass


class Harvest(LotLog):
    pass


class Observation(Log):
    pass


class Maintenance(Log):
    pass


class Purchase(MoneyLog):

    @property
    def seller(self) -> str:
        return self.attr('seller', str)


class Birth(Log):
    pass


class Medical(Log):
    pass


class Sale(Log):

    @property
    def invoice_number(self) -> str:
        return self.attr('invoice_number', str)

    @property
    def customer(self) -> str:
        return self.attr('customer', str)


class Activity(Log):
    pass


class SoilTest(Log):

    @property
    def soil_lab(self) -> str:
        return self.attr('soil_lab', str)

    @property
    def soil_names(self) -> List[Soil]:
        return []
